/******************************************************************************
 * PURPOSE: MapProjectionsInfo.c - Define functions for getting map projections
 *          information useful for building a user interface.
 * NOTES:   
 * HISTORY: 06/1995, Todd Plessel, EPA/MMTSI, Created stubs.
 *          09/1995, Mark Bolstad, EPA/MMTSI, Implemented.
 *          Required source-code control string:
 *          "%W% %P% %G% %U%"
 *****************************************************************************/

/*=============================== INCLUDES ==================================*/

#include <float.h>              /* For DBL_MIN, DBL_MAX. */

#include <Assertions.h>         /* For PRE(), POST(), etc.      */
#include <MapProjections.h>     /* For NUMBER_OF_ELLIPSES, etc. */
#include <MapProjectionsInfo.h> /* For public functions.        */

/*=============================== MACROS ===================================*/

/* FIX: these are unused: */

#define INVPROJ 0
#define FORPROJ 1

/*============================ PRIVATE VARIABLES ============================*/


static const ParameterInfo ellipseParams[NUMBER_OF_ELL_PARAMETERS] =
{
  { "A",  "a",  6378206.4, { 6378206.4, DBL_MAX } },
  { "B",  "b",  6356583.8, { 6356583.8, DBL_MAX } },
  { "RF", "rf", 0.0,       { 0.0,       DBL_MAX } },
  { "ES", "es", 0.0,       { 0.0,       DBL_MAX } }
};


static const EllipseInfo ellipseData[NUMBER_OF_ELLIPSES] =
{
  { "USE PARAMETERS",
    "none",     {1,1,1,1}, { 6371230.0, 6371230.0,       0.0,          0.0 } },
  { "MERIT 1983",
    "MERIT",    {1,0,1,0}, { 6378137.0,   0.0,         298.257,        0.0 } },
  { "SGS 85",
    "SGS85",    {1,0,1,0}, { 6378136.0,   0.0,         298.257,        0.0 } },
  { "GRS 1980(IUGG, 1980)",
    "GRS80",    {1,0,1,0}, { 6378137.0,   0.0,         298.257222101,  0.0 } },
  { "IAU 1976",
    "IAU76",    {1,0,1,0}, { 6378140.0,   0.0,         298.257,        0.0 } },
  { "Airy 1830",
    "airy",     {1,1,0,0}, { 6377563.396, 6356256.910, 0.0,            0.0 } },
  { "Appl. Physics. 1965",
    "APL4.9",   {1,0,1,0}, { 6378137.0,   0.0,         298.25,         0.0 } },
  { "Naval Weapons Lab., 1965",
    "NWL9D",    {1,0,1,0}, { 6378145.0,   0.0,         298.25,         0.0 } },
  { "Modified Airy",
    "mod_airy", {1,1,0,0}, { 6377340.189, 6356034.446,  0.0,           0.0 } },
  { "Andrae 1876 (Den., Iclnd.)",
    "andrae",   {1,0,1,0}, { 6377104.43,  0.0,          300.0,         0.0 } },
  { "Australian Natl & S. Amer. 1969",
    "aust_SA",  {1,0,1,0}, { 6378160.0,   0.0,          298.25,        0.0 } },
  { "GRS 67(IUGG 1967)",
    "GRS67",    {1,0,1,0}, { 6378160.0,   0.0,          247.247167,    0.0 } },
  { "Bessel 1841",
    "bessel",   {1,0,1,0}, { 6377397.155, 0.0,          299.1528128,   0.0 } },
  { "Bessel 1841 (Namibia)",
    "bess_nam", {1,0,1,0}, { 6377483.865, 0.0,          299.1528128,   0.0 } },
  { "Clarke 1866",
    "clrk66",   {1,1,0,0}, { 6378206.4,   6356583.8,    0.0,           0.0 } },
  { "Clarke 1880 mod.",
    "clrk80",   {1,0,1,0}, { 6378249.145, 0.0,          293.4663,      0.0 } },
  { "Comm. des Poids et Mesures 1799",
    "CPM",      {1,0,1,0}, { 6375738.7,   0.0,          334.29,        0.0 } },
  { "Delambre 1810 (Belgium)",
    "delmbr",   {1,0,1,0}, { 6376428.,    0.0,          311.5,         0.0 } },
  { "Engelis 1985",
    "engelis",  {1,0,1,0}, { 6378136.05,  0.0,          298.2566,      0.0 } },
  { "Everest 1830",
    "evrst30",  {1,0,1,0}, { 6377276.345, 0.0,          300.8017,      0.0 } },
  { "Everest 1948",
    "evrst48",  {1,0,1,0}, { 6377304.063, 0.0,          300.8017,      0.0 } },
  { "Everest 1956",
    "evrst56",  {1,0,1,0}, { 6377301.243, 0.0,          300.8017,      0.0 } },
  { "Everest 1969",
    "evrst69",  {1,0,1,0}, { 6377295.664, 0.0,          300.8017,      0.0 } },
  { "Everest (Sabah & Sarawak)",
    "evrstSS",  {1,0,1,0}, { 6377298.556, 0.0,          300.8017,      0.0 } },
  { "Fischer (Mercury Datum) 1960",
    "fschr60",  {1,0,1,0}, { 6378166.,    0.0,          298.3,         0.0 } },
  { "Modified Fischer 1960",
    "fschr60m", {1,0,1,0}, { 6378155.,    0.0,          298.3,         0.0 } },
  { "Fischer 1968",
    "fschr68",  {1,0,1,0}, { 6378150.,    0.0,          298.3,         0.0 } },
  { "Helmert 1906",
    "helmert",  {1,0,1,0}, { 6378200.,    0.0,          298.3,         0.0 } },
  { "Hough",
    "hough",    {1,0,1,0}, { 6378270.0,   0.0,          297.,          0.0 } },
  { "International 1909 (Hayford)",
    "intl",     {1,0,1,0}, { 6378388.0,   0.0,          297.,          0.0 } },
  { "Krassovsky, 1942",
    "krass",    {1,0,1,0}, { 6378245.0,   0.0,          298.3,         0.0 } },
  { "Kaula 1961",
    "kaula",    {1,0,1,0}, { 6378163.,    0.0,          298.24,        0.0 } },
  { "Lerch 1979",
    "lerch",    {1,0,1,0}, { 6378139.,    0.0,          298.257,       0.0 } },
  { "Maupertius 1738",
    "mprts",    {1,0,1,0}, { 6397300.,    0.0,          191.,          0.0 } },
  { "New International 1967",
    "new_intl", {1,1,0,0}, { 6378157.5,   6356772.2,    0.0,           0.0 } },
  { "Plessis 1817 (France)",
    "plessis",  {1,1,0,0}, { 6376523.,    6355863.,     0.0,           0.0 } },
  { "Southeast Asia",
    "SEasia",   {1,1,0,0}, { 6378155.0,   6356773.3205, 0.0,           0.0 } },
  { "Walbeck",
    "walbeck",  {1,1,0,0}, { 6376896.0,   6355834.8467, 0.0,           0.0 } },
  { "WGS 60",
    "WGS60",    {1,0,1,0}, { 6378165.0,   0.0,          298.3,         0.0 } },
  { "WGS 66",
    "WGS66",    {1,0,1,0}, { 6378145.0,   0.0,          298.25,        0.0 } },
  { "WGS 72",
    "WGS72",    {1,0,1,0}, { 6378135.0,   0.0,          298.26,        0.0 } },
  { "WGS 84",
    "WGS84",    {1,0,1,0}, { 6378137.0,   0.0,          298.257223563, 0.0 } }
};


static const ProjectionInfo projectionData[NUMBER_OF_PROJECTIONS] =
{
  { "Albers Equal Area", "aea",
    1,
    /* 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27*/
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0 }},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.4, {-90.0, 90.0}}}}},
  { "Azimuthal equidistant", "aeqd",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0", 29.5, {-90.0, 90.0}}}}},
  { "Airy", "airy",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "One Hemisphere", "no_cut", 0, {0, 1}}}}},
  { "Aitoff", "aitoff",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Modified Stereographics of Alaska", "alsk",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-152.0,{-152.1,-151.9}},
      {"Central Parallel", "lat_0", 64.0, {63.9, 64.1}}}}},
  { "Apian Globular I", "apian",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "August Epicycloidal", "august",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Bacon Globular", "bacon",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Bipolar Conic of Western Hemisphere", "bipc",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Boggs Eumorphic", "boggs",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Bonne (Werner lat_1 = 90)", "bonne",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}}}}},
  { "Cassini", "cass",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Central Cylindrical", "cc",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Cylindrical Equal Area", "cea",
    1,
    {{-1,-1, 0, 1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Lat of True Scale", "lat_ts", 0.0, {-90.0, 90.0}}}}},
  { "Chamberlin Trimetric", "chamb",
    1,
    {{-1,-1, 0, 1, 6, 7,-1,-1,-1, 4, 5,-1, 2, 3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Min Longitude","lon_1", -105.0, {-180.0, 180.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}},
      { "Max Longitude", "lon_2", -60.0, {-180.0, 180.0}},
      { "Chamb Latitude", "lat_3", 30.0, {-90.0, 90.0}},
      { "Chamb Longitude", "lon_3", -90.0, {-180.0, 180.0}}}}},
  { "Collignon", "collg",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Craster Parabolic (Putnins P4)", "crast",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Denoyer Semi-Elliptical", "denoy",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Eckert I", "eck1",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Eckert II", "eck2",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Eckert III", "eck3",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Eckert IV", "eck4",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Eckert V", "eck5",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Eckert VI", "eck6",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Equidistant Cylindrical (Plate Caree)", "eqc",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0", 0.0, {-90.0, 90.0}}}}},
  { "Equidistant Conic", "eqdc",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1, 2, 3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.4, {-90.0, 90.0}}}}},
  { "Euler", "euler",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1, 2, 3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.4, {-90.0, 90.0}}}}},
  { "Fahey", "fahey",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Foucaut", "fouc",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Foucaut Sinusoidal", "fouc_s",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Gall (Stereographic)", "gall",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Ginsburg VIII (Tsniigaik)", "gins8",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "General Sinusoidal Series", "gn_sinu",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 2, 3,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "m", "m", 1.0, {0.0,1.0}},
      { "n", "n", 1.0, {1.0,10.0}}}}},
  { "Gnomonic", "gnom",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Goode Homolosine", "goode",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Modified Stererographics of 48 U.S.", "gs48",
    1,
    {{-1,-1,-1, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-120.0,{-120.1,-119.9}},
      {"Central Parallel", "lat_0", 45.0, {44.9, 45.1}}}}},
  { "Modified Stererographics of 50 U.S.", "gs50",
    1,
    {{-1,-1,-1, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-120.0,{-120.1,-119.9}},
      {"Central Parallel", "lat_0", 45.0, {44.9, 45.1}}}}},
  { "Hammer & eckert-Greifendorff", "hammer",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Meridian Ratio", "W", 0.5, {0.0, DBL_MAX}},
      { "M", "M", 1.0, {0.0, DBL_MAX}}}}},
  { "Hatano Asymmetric Equal Area", "hatano",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "International Map of the World Polyconic", "imw_p",
    2,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
     {-1,-1, 0, 1,-1,-1,-1,-1,-1, 3,-1,-1, 2, 4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}}},
     {{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}},
      { "Min Longitude","lon_1", -60.0, {-180.0, 180.0}}}}},
  { "Kavraisky V", "kav5",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Kavraisky VII", "kav7",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
#if 0
  { "Laborde", "labrd",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0", 46.0,{45.9, 46.1}},
      {"Central Parallel", "lat_0", -18.0, {-18.1, -17.9}}}}},
#endif
  { "Lambert azimuthal Equal Area", "laea",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Lagrange", "lagrng",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Meridian Ratio", "W", 1.0, {0.0, DBL_MAX}}}}},
  { "Larrivee", "larr",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Laskowski", "lask",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Lambert Conformal Conic", "lcc",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}}}}},
  { "Lambert Equal Area Conic", "leac",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1, 3,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Southern Hemi", "south", 0, {0, 1}}}}},
  { "Lee Oblated Stereographic", "lee_os",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0", -165.0,{-165.1, -164.9}},
      {"Central Parallel", "lat_0", -10.0, {-10.1, -9.9}}}}},
  { "Loximuthal", "loxim",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Space Oblique for Landsat", "lsat",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 3,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Satellite Number", "lsat", 1, {1,5}},
      { "Path Number", "path", 1, {1,251}}}}},
  { "McBryde-Thomas Flat-polar Sine (No. 1)", "mbt_s",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "McBryde-Thomas Flat-polar Sine (No. 2)", "mbt_fps",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "McBryde-Thomas Flat-polar Parabolic", "mbtfpp",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "McBryde-Thomas Flat-polar Quartic", "mbtfpq",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "McBryde-Thomas Flat-polar Sinusoidal", "mbtfps",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Mercator", "merc",
    1,
    {{-1,-1, 0, 1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Lat of True Scale", "lat_ts", 35.0, {-90.0, 90.0}}}}},
  { "Miller Oblated Stereographic", "mil_os",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0", 20.0,{ 19.9, 20.1 }},
      {"Central Parallel", "lat_0", 18.0, {17.9, 18.1}}}}},
  { "Miller Cylindrical", "mill",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
#if 0
  { "Modified Polyconic", "mpoly",
    1,
    {{-1,-1, 0, 1,-1,-1,-1, 4,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}},
      { "Lotsa", "lotsa", 0, {0,1}}}}},
#endif
  { "Mollweide","moll",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Murdoch I", "murd1",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}}}}},
  { "Murdoch II", "murd2",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}}}}},
  { "Murdoch III", "murd3",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}}}}},
  { "Nell", "nell",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Nell-Hammer", "nell_h",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Nicolosi Globular", "nicol",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
#if 0
  { "Near-sided Perspective", "nsper",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 2,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "View Pt Height", "h", 8.0, {0.0, DBL_MAX}}}}},
#endif
#if 0
  { " New Zealand Map Grid", "nzmg",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",173.0,{172.9, 173.1}},
      {"Central Parallel", "lat_0", -41.0, {-41.1, -40.9}}}}},
#endif
  { "Oblique Cylindrical Equal Area", "ocea",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 4, 5,-1, 2, 3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Min Longitude","lon_1", -105.0,{-180.0, 180.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}},
      { "Max Longitude", "lon_2", -60.0, {-180.0, 180.0}}}}},
  { "Oblated Equal Area", "oea",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 4,-1,-1,-1, 3, 2,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "n", "n", 1, {DBL_MIN, DBL_MAX}},
      { "m", "m", 1, {DBL_MIN, DBL_MAX}},
      { "Theta", "theta", 0.0, {-360.0, 360.0}}}}},
  { "Oblique Mercator", "omerc",
    2,
    {{ 5, 6, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 2, 4,-1,-1, 3,-1,-1,-1,-1,-1,-1,-1,-1,-1},
     {-1,-1, 0, 1,-1,-1,-1,-1,-1, 7, 8,-1, 5, 6, 2, 4,-1,-1, 3,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "No Rotation", "no_rot", 0, {0, 1}},
      { "Rotation Conversion", "rot_conv", 0, {0, 1}},
      { "No U Offset", "no_uoff", 0, {0, 1}},
      { "Azimuth", "alpha", 10.0, {-180.0, 180.0}},
      { "Central Lon", "lonc", -90.0, {-180.0, 180.0}}},
     {{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "No Rotation", "no_rot", 0, {0, 1}},
      { "Rotation Conversion", "rot_conv", 0, {0, 1}},
      { "No U Offset", "no_uoff", 0, {0, 1}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Min Longitude","lon_1", -105.0,{-180.0, 180.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}},
      { "Max Longitude", "lon_2", -60.0, {-180.0, 180.0}}}}},
  { "Ortelius Oval", "ortel",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Orthographic", "ortho",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      {"Central Parallel", "lat_0", 0.0, {-90.0, 90.0}}}}},
  { "Perspective Conic", "pconic",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}}}}},
  { "Polyconic (American)", "poly",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}}}}},
  { "Putnins P1", "putp1",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Putnins P2", "putp2",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Putnins P3", "putp3",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Putnins P3'", "putp3p",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Putnins P4'", "putp4p",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Putnins P5", "putp5",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Putnins P5'", "putp5p",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Putnins P6", "putp6",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Putnins P6'", "putp6p",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Quartic Authalic", "qua_aut",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Robinson", "robin",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Rectangular Polyconic", "rpoly",
    1,
    {{-1,-1, 0, 1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Lat of True Scale", "lat_ts", 0.0, {-90.0, 90.0}}}}},
  { "Sinusoidal (Sanson-Flamsteed)", "sinu",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Stereographic", "stere",
    1,
    {{-1,-1, 0, 1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Lat of True Scale", "lat_ts", 0.0, {-90.0, 90.0}}}}},
  { "Transverse Central Cylindrical", "tcc",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Transverse Cylindrical Equal Area", "tcea",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Tissot", "tissot",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}}}}},
  { "Transverse Mercator", "tmerc",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Two Point Equidistant", "tpeqd",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 4, 5,-1, 2, 3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      { "Min Longitude","lon_1", -105.0,{-180.0, 180.0}},
      { "Max Latitude", "lat_2", 45.5, {-90.0, 90.0}},
      { "Max Longitude", "lon_2", -60.0, {-180.0, 180.0}}}}},
#if 0
  { "Tilted Perspective", "tpers",
    1,
    {{ 1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 2, 3,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Azimuth", "azi", 0.0, {-180.0, 180.0}},
      { "Tilt", "tilt", 0.0, {-180.0, 180.0}},
      { "View Pt Height", "h", 1.0, {0.0, DBL_MAX}}}}},
#endif
  { "Universal Polar Stereographic", "ups",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "Southern Hemi", "south", 0, {0, 1}}}}},
  { "Urmaev V", "urm5",
    1,
    {{ 3,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 2, 3}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      { "n", "n", 1.0, {-DBL_MAX,DBL_MAX}},
      {"q", "q", 1.0, {-DBL_MAX,DBL_MAX}},
      {"Azimuth", "alpha", 0.0, {-180.0, 180.0}}}}},
  { "Urmaev Flat-polar Sinusoidal", "urmfps",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 2,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      {"n", "n", 1.0, {1.0,1.0}}}}},
  { "Universal Transverse Mercator (UTM)", "utm",
    2,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1},
     {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 1,-1,-1,-1, 0,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      {"Southern Hemi", "south", 0, {0, 1}}},
     {{"Zone", "zone", 1, {1,60}},
      {"Southern Hemi", "south", 0, {0, 1}}}}},
  { "Van Der Grinten I", "vandg",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Van Der Grinten II", "vandg2",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Van Der Grinten III", "vandg3",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Van Der Grinten IV", "vandg4",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Vitkovsky I", "vitk1",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1, 3,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      {"Min Latitude", "lat_1", 29.5, {-90.0, 90.0}},
      {"Max Latitude", "lat_2", 45.5, {-90.0, 90.0}}}}},
  { "Wagner I (Kavraisky VI)", "wag1",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Wagner II", "wag2",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Wagner III", "wag3",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Wagner IV", "wag4",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Wagner V", "wag5",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Wagner VI", "wag6",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Wagner VII", "wag7",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Werenskiold I", "weren",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}},
  { "Winkel I", "wink1",
    1,
    {{-1,-1, 0, 1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      {"Lat of True Scale", "lat_ts", 0.0, {-90.0, 90.0}}}}},
  { "Winkel II", "wink2",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      {"Min Latitude", "lat_1", 29.5, {-90.0, 90.0}}}}},
  { "Winkel Tripel", "wintri",
    1,
    {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
    {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}},
      {"Min Latitude", "lat_1", 29.5, {-90.0, 90.0}}}}},
  { "Lat-Lon", "none",
   1,
   {{-1,-1, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
   {{{"Central Meridian","lon_0",-90.0,{-180.0,180.0}},
      { "Central Parallel", "lat_0",0.0,{-90.0,90.0}}}}}};


/*=========================== PRIVATE FUNCTIONS =============================*/


/******************************************************************************
 * PURPOSE: lookup - Map the mapProjection parameter array into the approriate
 *                   index into the parameters structure.
 * INPUTS:  int projection  The projection.
 *          int parameter   The parameter.
 *          int alternative The alternative.
 * OUTPUTS: None
 * RETURNS: The index in the parameters table.
 * NOTES:
 *****************************************************************************/

static int lookup( int projection, int parameter, int alternative )
{
  PRE3( IN_RANGE( projection,  0, NUMBER_OF_PROJECTIONS         - 1 ),
        IN_RANGE( parameter,   0, NUMBER_OF_PARAMETERS          - 1 ),
        IN_RANGE( alternative, 0, numAlternatives( projection ) - 1 ) );

  return projectionData[ projection ].enumMapping[ alternative ][ parameter ];
}


/*=========================== PUBLIC FUNCTIONS ==============================*/


/******************************************************************************
 * PURPOSE: numAlternatives - Get the number of alternatives for a projection.
 * INPUTS:  int projection  The projection.
 * OUTPUTS: None
 * RETURNS: int The number of alternatives for this projection.
 * NOTES:
 *****************************************************************************/

int numAlternatives( int projection )
{
  PRE( IN_RANGE( projection, 0, NUMBER_OF_PROJECTIONS - 1 ) );

  return projectionData[ projection ].numAlternatives;
}


/******************************************************************************
 * PURPOSE: getMapProjectionParamTag - Get the parameter tag used in the
 *                                     projection library for a given set of
 *                                     projection, parameter and alternative.
 * INPUTS:  int projection  The projection.
 *          int parameter   The parameter.
 *          int alternative The alternative.
 * OUTPUTS: None
 * RETURNS: const char*     The string tag.
 * NOTES:
 *****************************************************************************/

const char* getMapProjectionParamTag( int projection, int parameter,
                                      int alternative )
{
  PRE4( IN_RANGE( projection,  0, NUMBER_OF_PROJECTIONS         - 1 ),
        IN_RANGE( parameter,   0, NUMBER_OF_PARAMETERS          - 1 ),
        IN_RANGE( alternative, 0, numAlternatives( projection ) - 1 ),
        lookup( projection, parameter, alternative ) != -1 );

  int i = lookup( projection, parameter, alternative );
  return projectionData[ projection ].parameters[ alternative ][ i ].projName;
}


/******************************************************************************
 * PURPOSE: getEllipseParamTag - Get the ellipse tag used in the proj. library.
 * INPUTS:  int ellipse  The ellipse.
 * OUTPUTS: None
 * RETURNS: const char*  The string tag.
 * NOTES:
 *****************************************************************************/

const char* getEllipseParamTag( int ellipse )
{
  PRE( IN_RANGE( ellipse, 0, NUMBER_OF_ELL_PARAMETERS - 1 ) );

  return ellipseParams[ellipse].projName;
}


/******************************************************************************
 * PURPOSE: getEllipseProjName - Get the PROJ Library name of an ellipse.
 * INPUTS:  int ellispe  The ellispe.
 * OUTPUTS: None
 * RETURNS: const char* The parameter name.
 * NOTES:
 *****************************************************************************/

const char* getEllipseProjName( int ellipse )
{
  PRE( IN_RANGE( ellipse, 0, NUMBER_OF_ELLIPSES - 1 ) );

  return ellipseData[ ellipse ].projName;
}


/******************************************************************************
 * PURPOSE: getEllipseName - Get the name of an ellipse.
 * INPUTS:  int ellispe  The ellispe.
 * OUTPUTS: None
 * RETURNS: const char*  The string tag.
 * NOTES:
 *****************************************************************************/

const char* getEllipseName( int ellipse )
{
  PRE( IN_RANGE( ellipse, 0, NUMBER_OF_ELLIPSES - 1 ) );

  return ellipseData[ ellipse ].name;
}


/******************************************************************************
 * PURPOSE: getEllipseParamIniter - Get the initial value for an ellipse.
 * INPUTS:  int ellipse     The ellipse.
 *          int parameter   The ellipse parameter.
 * OUTPUTS: None
 * RETURNS: double  The initializer.
 * NOTES:
 *****************************************************************************/

double getEllipseParamIniter( int ellipse, int parameter )
{
  PRE2( IN_RANGE( ellipse,    0, NUMBER_OF_ELLIPSES - 1 ),
        IN_RANGE( parameter,  0, NUMBER_OF_ELL_PARAMETERS - 1 ) );

  return ellipseData[ ellipse ].initVal[ parameter ];
}


/******************************************************************************
 * PURPOSE: getMapProjectionName - Get the name of a projection.
 * INPUTS:  int projection  The projection.
 * OUTPUTS: None
 * RETURNS: const char*  The string tag.
 * NOTES:
 *****************************************************************************/

const char* getMapProjectionName( int projection )
{
  PRE( IN_RANGE( projection, 0, NUMBER_OF_PROJECTIONS - 1 ) );

  return projectionData[ projection ].name;
}


/******************************************************************************
 * PURPOSE: getMapProjectionProjName - Get the name of a projection.
 * INPUTS:  int projection  The projection.
 * OUTPUTS: None
 * RETURNS: const char*  The string tag.
 * NOTES:
 *****************************************************************************/

const char* getMapProjectionProjName( int projection )
{
  PRE( IN_RANGE( projection, 0, NUMBER_OF_PROJECTIONS - 1 ) );

  return projectionData[ projection ].projName;
}


/******************************************************************************
 * PURPOSE: getMapProjectionVisibility - Get the 'visibility' or validity
 *          values
 *                                       for a projection-alternative pair.
 * INPUTS:  int projection  The projection.
 *          int alternative The alternative.
 * OUTPUTS: None
 * RETURNS: const int*  The array of validity flags.
 * NOTES:
 *****************************************************************************/

const int* getMapProjectionVisibility( int projection, int alternative )
{
  PRE2( IN_RANGE( projection,  0, NUMBER_OF_PROJECTIONS         - 1 ),
        IN_RANGE( alternative, 0, numAlternatives( projection ) - 1 ) );

  return projectionData[ projection ].enumMapping[ alternative ];
}


/******************************************************************************
 * PURPOSE: getMapProjectionParamIniter - Get the initial value for a
 *                                        projection, parameter, alternative
 *                                        triple.
 * INPUTS:  int projection  The projection.
 *          int parameter   The parameter.
 *          int alternative The alternative.
 * OUTPUTS: None
 * RETURNS: double  The initializer.
 * NOTES:
 *****************************************************************************/

double getMapProjectionParamIniter( int projection, int parameter,
                                    int alternative )
{
  PRE4( IN_RANGE( projection,  0, NUMBER_OF_PROJECTIONS         - 1 ),
        IN_RANGE( parameter,   0, NUMBER_OF_PARAMETERS          - 1 ),
        IN_RANGE( alternative, 0, numAlternatives( projection ) - 1 ),
        lookup( projection, parameter, alternative ) != -1 );

  int i = lookup( projection, parameter, alternative );
  return projectionData[ projection ].parameters[ alternative ][ i ].initVal;
}


/******************************************************************************
 * PURPOSE: getParamameterExtrema - Get the 'extrema' or bounds for a
 *                                  projection, parameter, alternative triple.
 * INPUTS:  int projection  The projection.
 *          int parameter   The parameter.
 *          int alternative The alternative.
 * OUTPUTS: None
 * RETURNS: const double*   Pointer to the extrema[2].
 * NOTES:
 *****************************************************************************/

const double* getParameterExtrema( int projection, int parameter,
                                   int alternative )
{
  PRE4( IN_RANGE( projection,  0, NUMBER_OF_PROJECTIONS         - 1 ),
        IN_RANGE( parameter,   0, NUMBER_OF_PARAMETERS          - 1 ),
        IN_RANGE( alternative, 0, numAlternatives( projection ) - 1 ),
        lookup( projection, parameter, alternative ) != -1 );

  int i = lookup( projection, parameter, alternative );

  return projectionData[ projection ].parameters[ alternative ][ i ].extrema;
}

